@@ -8,6 +8,7 @@ from ticdat.utils import ForeignKey, ForeignKeyMapping, TypeDictionary, verify,
 from ticdat.utils import lupish, deep_freeze, containerish, FrozenDict, safe_apply, stringish
 import ticdat.pandatio as pandatio
 from itertools import count
+import inspect
 import math
 try:
     from pandas import isnull
@@ -23,6 +24,11 @@ except:
 
 pd, DataFrame = utils.pd, utils.DataFrame # if pandas not installed will be falsey
 
+def _safe_any(s):
+    if len(s) == 0:
+        return False
+    return s.any()
+
 def _is_last_rows_all_nan(df, last_rows):
     assert last_rows > 0
     # quick last row check to make faster
@@ -107,6 +113,7 @@ class PanDatFactory(object):
                 "infinity_io_flag": self.infinity_io_flag,
                 "xlsx_trailing_empty_rows": self.xlsx_trailing_empty_rows,
                 "duplicates_ticdat_init": self.duplicates_ticdat_init,
+                "automunge_multitype_fields": self.automunge_multitype_fields,
                 "tooltips": utils.make_tooltips_dict_json_friendly(self.tooltips)}
     @staticmethod
     def create_from_full_schema(full_schema):
@@ -123,7 +130,8 @@ class PanDatFactory(object):
         old_schema = {"tables_fields", "foreign_keys", "default_values", "data_types"}
         verify(dictish(full_schema) and set(full_schema).issuperset(old_schema) and set(full_schema) in
                utils.all_subsets(old_schema.union({"parameters", "infinity_io_flag", "xlsx_trailing_empty_rows",
-                                                   "duplicates_ticdat_init", "tooltips"})),
+                                                   "duplicates_ticdat_init", "tooltips",
+                                                   "automunge_multitype_fields"})),
                "full_schema should be the result of calling schema(True) for some PanDatFactory")
         fks = full_schema["foreign_keys"]
         verify( (not fks) or (lupish(fks) and all(lupish(_) and len(_) >= 3 for _ in fks)),
@@ -162,12 +170,14 @@ class PanDatFactory(object):
             rtn.set_xlsx_trailing_empty_rows(full_schema["xlsx_trailing_empty_rows"])
         if "duplicates_ticdat_init" in full_schema:
             rtn.set_duplicates_ticdat_init(full_schema["duplicates_ticdat_init"])
+        if "automunge_multitype_fields" in full_schema:
+            rtn.set_automunge_multitype_fields(full_schema["automunge_multitype_fields"])
         if "tooltips" in full_schema:
             for t, tip_dict in full_schema["tooltips"].items():
                 for f, tip in tip_dict.items():
                     rtn.set_tooltip(t, f, tip)
         return rtn
-    def clone(self, table_restrictions=None, clone_factory=None):
+    def clone(self, table_restrictions=None, clone_factory=None, convert_dat=None):
         """
 
         clones the PanDatFactory
@@ -179,7 +189,12 @@ class PanDatFactory(object):
         :param clone_factory : optional. Defaults to PanDatFactory. Can also be TicDatFactory.  Can also be a function,
                                in which case it should behave similarly to create_from_full_schema.
                                If clone_factory=TicDatFactory, the row predicates that use predicate_kwargs_maker
-                               won't be copied over.
+                               will only be copied over if a convert_dat argument is provided.
+
+        :param convert_dat : optional. A function that converts the dat object before it is passed to any
+                                       predicate_kwargs_maker functions. This function will be applied to the
+                                       predicate_kwargs_maker references from row predicates copied over as part of this
+                                       clone, and not to row predicates added subsequently.
 
         :return: a clone of the PanDatFactory. Returned object will be based on clone_factory, if provided.
 
@@ -190,7 +205,7 @@ class PanDatFactory(object):
         """
         clone_factory = clone_factory or PanDatFactory
         from ticdat import TicDatFactory
-        no_copy_predicate_kwargs_maker = clone_factory == TicDatFactory
+        no_copy_predicate_kwargs_maker = (clone_factory == TicDatFactory and not convert_dat)
         if hasattr(clone_factory, "create_from_full_schema"):
             clone_factory = clone_factory.create_from_full_schema
         full_schema = utils.clone_a_anchillary_info_schema(self.schema(include_ancillary_info=True),
@@ -203,6 +218,13 @@ class PanDatFactory(object):
                         rtn.add_data_row_predicate(tbl, predicate=rpi.predicate, predicate_name=pn,
                                                    predicate_kwargs_maker=rpi.predicate_kwargs_maker,
                                                    predicate_failure_response=rpi.predicate_failure_response)
+        if convert_dat: # this function is effectively a constructor so _ reference is ok
+            assert callable(convert_dat) and len(inspect.getfullargspec(convert_dat).args) >= 1
+            rtn._convert_dat[:] = [convert_dat,
+                                   {(tbl, pn) for tbl, pns in rtn._data_row_predicates.items()
+                                    for pn, rpi in pns.items() if rpi.predicate_kwargs_maker}]
+        else:
+            rtn._convert_dat[:] = self._convert_dat[:]
         return rtn
     def clone_add_a_table(self, table, pk_fields, df_fields):
         '''
@@ -333,6 +355,24 @@ class PanDatFactory(object):
         verify(value in ["prune", "ignore"], f"bad value {value}")
         self._xlsx_trailing_empty_rows[0] = value
     @property
+    def automunge_multitype_fields(self):
+        """
+        see __doc__ for set_automunge_multitype_fields
+        :return:
+        """
+        return self._automunge_multitype_fields[0]
+    def set_automunge_multitype_fields(self, value):
+        """
+        Set the automunge_multitype_fields for the PanDatFactory. Choices are: True, False
+        Default is True. Prior to 0.2.28, there was no automunging of multitype fields, so if you want to
+        retain that behavior with the 0.2.28 release, you will need to call this function with value=False.
+        See https://github.com/ticdat/ticdat/wiki/ticdat-automunging for more details.
+        :param value:
+        :return:
+        """
+        verify(value in [True, False],  f"bad value {value}")
+        self._automunge_multitype_fields[0] = value
+    @property
     def infinity_io_flag(self):
         """
         see __doc__ for set_infinity_io_flag
@@ -407,7 +447,21 @@ class PanDatFactory(object):
         :param json_read: special 'None'->None override needed for pandas json reader
         '''
         assert push_parameters_to_be_valid or not json_read, "json_read should always push_parameters_to_be_valid"
-        apply = utils.faster_df_apply
+        def _multitype_cell_adj(x):
+            _x = safe_apply(float)(x)
+            return x if _x is None else _x
+        def _datetime_cell_adj(x):
+            _x = utils.dateutil_adjuster(x)
+            return x if _x is None else _x
+        def _json_none_cell_adj(x):
+            if utils.stringish(x) and x.lower() == "none":
+                return None
+            return x
+        def _inf_cell_adj(x):
+            if not utils.numericish(x):
+                return x
+            return float("inf") if x >= self.infinity_io_flag else \
+                (-float("inf") if x <= -self.infinity_io_flag else x)
         for t in set(self.all_tables).difference(["parameters"]): # parameters table is handled differently
             df = getattr(dat, t)
             all_fields = tuple(self.primary_key_fields.get(t, ()) + self.data_fields.get(t, ()))
@@ -417,32 +471,24 @@ class PanDatFactory(object):
                     for f in all_fields:
                         if utils.numericish(row[f]) and abs(row[f]) >= self.infinity_io_flag:
                             fields_w_issues.add(f)
-                apply(df, find_fields_w_issues)
+                utils.faster_df_apply(df, find_fields_w_issues)
                 for f in fields_w_issues:
-                    fixme = apply(df, lambda row: utils.numericish(row[f]) and row[f] >= self.infinity_io_flag)
-                    if fixme.any():
-                        df.loc[fixme, f] = float("inf")
-                    fixme = apply(df, lambda row: utils.numericish(row[f]) and row[f] <= -self.infinity_io_flag)
-                    if fixme.any():
-                        df.loc[fixme, f] = -float("inf")
+                    df[f] = df[f].apply(_inf_cell_adj)
             for f in all_fields:
+                dt = self.data_types.get(t, {}).get(f, None)
+                if dt and not dt.datetime and dt.strings_allowed and dt.number_allowed and \
+                    self.automunge_multitype_fields:
+                    df[f] = df[f].apply(_multitype_cell_adj)
                 if not utils.numericish(self.infinity_io_flag) and utils.numericish(self._none_as_infinity_bias(t, f)):
                     assert self.infinity_io_flag is None
-                    df[f].fillna(value=self._none_as_infinity_bias(t, f) * float("inf"), inplace=True)
-                dt = self.data_types.get(t, {}).get(f, None)
+                    df[f] = df[f].apply(lambda x: self._none_as_infinity_bias(t, f) * float("inf")
+                                        if pd.isnull(x) else x)
                 if dt and dt.datetime:
-                    def fixed_row(row):
-                        new_row_f = utils.dateutil_adjuster(row[f])
-                        return new_row_f if new_row_f is not None else row[f]
-                    df[f] = apply(df, fixed_row)
+                    df[f] = df[f].apply(_datetime_cell_adj)
                 if json_read and self._dtypes_for_pandas_read(t).get(f) == str:
                     assert dt, "assumed because _dtypes_for_pandas_read result"
                     if dt.nullable:
-                        def fixed_row(row):
-                            if utils.stringish(row[f]) and row[f].lower() == "none":
-                                return None
-                            return row[f]
-                        df[f] = apply(df, fixed_row)
+                        df[f] = df[f].apply(_json_none_cell_adj)
 
         # this is the logic that is used in lieu of infinity_io_flag logic for the parameters table
         # it is predicated on the assumption that the parameters table will be serialized to a string/string table
@@ -506,10 +552,10 @@ class PanDatFactory(object):
                 apply(df, find_fields_w_issues)
                 for f in fields_w_issues:
                     fixme = apply(df, lambda row: utils.numericish(row[f]) and row[f] >= self.infinity_io_flag)
-                    if fixme.any():
+                    if _safe_any(fixme):
                         df.loc[fixme, f] = self.infinity_io_flag
                     fixme = apply(df, lambda row: utils.numericish(row[f]) and row[f] <= -self.infinity_io_flag)
-                    if fixme.any():
+                    if _safe_any(fixme):
                         df.loc[fixme, f] = -self.infinity_io_flag
             else:
                 all_fields = tuple(f for f in all_fields if utils.numericish(self._none_as_infinity_bias(t, f)))
@@ -521,7 +567,7 @@ class PanDatFactory(object):
                 for f in fields_w_issues:
                     assert self.infinity_io_flag is None
                     fixme = apply(df, lambda row: row[f] == float("inf") * self._none_as_infinity_bias(t, f))
-                    if fixme.any():
+                    if _safe_any(fixme):
                         df.loc[fixme, f] = None
         return rtn
     def set_data_type(self, table, field, number_allowed = True,
@@ -607,9 +653,9 @@ class PanDatFactory(object):
 
         !!! NB!!!!
 
-        **pandas will render None as nan.**
+        **pandas will typically render NULL as nan. In rare cases, it uses None.**
 
-        **Don't check for None in your predicate functions, use pandas.isnull instead**
+        **Don't check for None (or nan) in your predicate functions. Use pandas.isnull**
 
         !!!!!!!!!!
 
@@ -922,8 +968,9 @@ class PanDatFactory(object):
         self._infinity_io_flag = ["N/A"]
         self._xlsx_trailing_empty_rows = ["prune"]
         self._duplicates_ticdat_init = ["assert"]
+        self._automunge_multitype_fields = [True]
         self._none_as_infinity_bias_cache = {}
-
+        self._convert_dat = []
 
         self.all_tables = frozenset(init_fields)
         superself = self
@@ -953,8 +1000,9 @@ class PanDatFactory(object):
                     if pks and not set(pks).intersection(df.columns) and \
                        set(pks) == utils.safe_apply(lambda: set(df.index.names))():
                         df.reset_index(drop=False, inplace=True)
-                    if list(df.columns) == list(range(len(df.columns))) and \
-                       len(df.columns) >= len(superself._all_fields(t)):
+                    if list(df.columns) == list(range(len(df.columns))):
+                        verify(len(df.columns) >= len(superself._all_fields(t)),
+                               f"{t} cannot be treated as a PanDat table : insufficient number of columns.")
                         df.rename(columns={f1:f2 for f1, f2 in zip(df.columns, superself._all_fields(t))},
                                   inplace=True)
                     if list(df.columns) != list(range(len(df.columns))):
@@ -982,31 +1030,41 @@ class PanDatFactory(object):
         self.json = pandatio.JsonPanFactory(self)
         self.pgsql = PostgresPanFactory(self)
 
-    def good_pan_dat_object(self, data_obj, bad_message_handler = lambda x : None):
+    def good_pan_dat_object(self, data_obj, bad_message_handler=None):
         """
         determines if an object is a valid PanDat object for this schema
 
         :param data_obj: the object to verify
 
         :param bad_message_handler: a call back function to receive description of any failure message
+                                    or None if you want fast fail with no diagnostics
 
         :return: True if the dataObj can be recognized as a PanDat data object. False otherwise.
         """
+        verify(bad_message_handler is None or callable(bad_message_handler),
+               "need bad_message_handler callable or None")
+        fast_fail = not bad_message_handler
+        bad_message_handler = bad_message_handler or (lambda x: None)
+
         verify(DataFrame and pd, "Need to install pandas")
+        df_found = {}
         for t in self.all_tables:
             if not hasattr(data_obj, t) :
                 bad_message_handler(t + " not an attribute.")
-                return False
-            if not isinstance(getattr(data_obj, t), DataFrame):
+                if fast_fail:
+                    return False
+            elif not isinstance(getattr(data_obj, t), DataFrame):
                 bad_message_handler(t + " is not a DataFrame")
-                return False
-        missing_fields = {(t, f) for t in self.all_tables for f in
+                if fast_fail:
+                    return False
+            else:
+                df_found[t] = getattr(data_obj, t)
+        missing_fields = {(t, f) for t in df_found for f in
                           self.primary_key_fields.get(t, ()) + self.data_fields.get(t, ())
-                          if f not in getattr(data_obj, t).columns}
+                          if f not in df_found[t].columns}
         if missing_fields:
-            bad_message_handler("The following are (table, field) pairs missing from the data.\n%s"%missing_fields)
-            return False
-        return True
+            bad_message_handler(f"The following are (table, field) pairs missing from the data.\n{missing_fields}")
+        return set(df_found) == set(self.all_tables) and not missing_fields
     def copy_pan_dat(self, pan_dat):
         """
         copies the tic_dat object into a new tic_dat object
@@ -1120,21 +1178,23 @@ class PanDatFactory(object):
         rtn = {}
         TableField = clt.namedtuple("TableField", ["table", "field"])
         number_failures = [0]
-        check_too_many = None
-        if max_failures < float("inf"):
-            def check_too_many(is_bad_row):
-                if is_bad_row:
-                    number_failures[0] += 1
-                    if number_failures[0] >= max_failures:
-                        return lambda row: False  # all future rows will be good (i.e. not bad, i.e. False)
         for table, type_row in self._true_data_types().items():
             _table = getattr(pan_dat, table)
             for field, data_type in type_row.items():
-                def bad_row(row):
-                    data = row[field]
+                def _bad_row(data):
                     return not data_type.valid_data(None if isnull(data) else data)
-                where_bad_rows = utils.faster_df_apply(_table, bad_row, trip_wire_check=check_too_many)
-                if where_bad_rows.any():
+                if max_failures == float("inf"):
+                    bad_row = _bad_row
+                else:
+                    def bad_row(data):
+                        if number_failures[0] >= max_failures:
+                            return False  # all future rows will be good (i.e. not bad, i.e. False)
+                        rtn = _bad_row(data)
+                        if rtn:
+                            number_failures[0] += 1
+                        return rtn
+                where_bad_rows = _table[field].apply(bad_row)
+                if _safe_any(where_bad_rows):
                     rtn[TableField(table, field)] = _table[where_bad_rows].copy() if as_table else where_bad_rows
                 if number_failures[0] >= max_failures:
                     return rtn
@@ -1186,9 +1246,14 @@ class PanDatFactory(object):
                 verify(self._true_data_types()[table][field].valid_data(value),
                        "The replacement value %s is not itself valid for %s : %s"%(value, table, field))
 
-        for (table, field), rows in replacements_needed.items() :
+        for table, field in replacements_needed:
             if (table, field) in real_replacements:
-                getattr(pan_dat, table).loc[rows, field] = real_replacements[table, field]
+                data_type = self._true_data_types()[table][field]
+                def bad_row(row):
+                    data = row[field]
+                    return not data_type.valid_data(None if isnull(data) else data)
+                getattr(pan_dat, table)[field] = utils.faster_df_apply(getattr(pan_dat, table), lambda row:
+                        real_replacements[table, field] if bad_row(row) else row[field])
         assert not set(self.find_data_type_failures(pan_dat)).intersection(real_replacements)
         return pan_dat
     def find_data_row_failures(self, pan_dat, as_table=True, exception_handling="__debug__",
@@ -1273,19 +1338,32 @@ class PanDatFactory(object):
         predicate_kwargs_maker_results = {}
         TPN = clt.namedtuple("TablePredicateName", ["table", "predicate_name"])
         PKEM = clt.namedtuple("PrimaryKeyErrorMessage", ["primary_key", "error_message"])
+        converted_dat = []
         for tbl, row_predicates in data_row_predicates.items():
             _table = getattr(pan_dat, tbl)
             for pn, rpi in row_predicates.items():
+                uses_convert = self._convert_dat and (tbl, pn) in self._convert_dat[1]
                 predicate_kwargs = {}
                 if rpi.predicate_kwargs_maker:
-                    if rpi.predicate_kwargs_maker not in predicate_kwargs_maker_results:
+                    if uses_convert and not converted_dat:
                         if exception_handling == "Handled as Failure":
                             try:
-                                _predicate_kwargs = rpi.predicate_kwargs_maker(pan_dat)
+                                converted_dat.append(self._convert_dat[0](pan_dat))
+                            except Exception as e:
+                                converted_dat.append(f"Exception<{e}>")
+                        else:
+                            converted_dat.append(self._convert_dat[0](pan_dat))
+                    if rpi.predicate_kwargs_maker not in predicate_kwargs_maker_results:
+                        __pan_dat = converted_dat[0] if uses_convert else pan_dat
+                        if uses_convert and isinstance(converted_dat[0], str):
+                            _predicate_kwargs = converted_dat[0]
+                        elif exception_handling == "Handled as Failure":
+                            try:
+                                _predicate_kwargs = rpi.predicate_kwargs_maker(__pan_dat)
                             except Exception as e:
                                 _predicate_kwargs = f"Exception<{e}>"
                         else:
-                            _predicate_kwargs = rpi.predicate_kwargs_maker(pan_dat)
+                            _predicate_kwargs = rpi.predicate_kwargs_maker(__pan_dat)
                         predicate_kwargs_maker_results[rpi.predicate_kwargs_maker] = _predicate_kwargs
                     predicate_kwargs = predicate_kwargs_maker_results[rpi.predicate_kwargs_maker]
                 if not isinstance(predicate_kwargs, dict):
@@ -1304,7 +1382,7 @@ class PanDatFactory(object):
                                   if exception_handling == "Unhandled" else (lambda row: not _p(row))
 
                         where_bad_rows = utils.faster_df_apply(_table, bad_row, trip_wire_check=check_too_many_bool)
-                        if where_bad_rows.any():
+                        if _safe_any(where_bad_rows):
                             rtn[TPN(tbl, pn)] = _table[where_bad_rows].copy() if as_table else where_bad_rows
                     else:
                         def _p(row):
@@ -1316,7 +1394,7 @@ class PanDatFactory(object):
                                     if exception_handling == "Unhandled" else (lambda row: _p(row))
                         predicate_result = utils.faster_df_apply(_table, predicate, trip_wire_check=check_too_many_msg)
                         where_bad_rows = predicate_result.apply(lambda x: x is not True)
-                        if where_bad_rows.any():
+                        if _safe_any(where_bad_rows):
                             if as_table:
                                 rtn[TPN(tbl, pn)] = _df = _table[where_bad_rows].copy()
                                 err_column = "Error Message"
@@ -1408,8 +1486,14 @@ class PanDatFactory(object):
             child.insert(0, magic_field*2, range(0, len(child)))
             parent.set_index(new_index, drop=True, inplace=True)
             child.set_index(new_index, drop=True, inplace=True)
-            joined = child.join(parent, rsuffix=magic_field)
-            bad_rows = set(joined[joined[magic_field] != True][magic_field*2])
+            try:
+                joined = child.join(parent, rsuffix=magic_field)
+            except TypeError:
+                joined = None # see ticdat issue 173
+            except:
+                raise
+            bad_rows = set(child[magic_field*2]) if joined is None else \
+                       set(joined[joined[magic_field] != True][magic_field*2])
             if bad_rows:
                 # I'm casting to list here because child is a copy that doesn't have the original index
                 # This is fixable, (i.e. we could return a Series with an index matching the original child table)
@@ -1484,9 +1568,44 @@ class PanDatFactory(object):
         for t in self.all_tables:
             if self.primary_key_fields.get(t):
                 dups = getattr(pan_dat, t).duplicated(list(self.primary_key_fields[t]), keep=keep)
-                if dups.any():
+                if _safe_any(dups):
                     rtn[t] = getattr(pan_dat, t)[list(dups)] if as_table else dups
         return rtn
+    def obfusimplify(self, pan_dat, table_prepends=utils.FrozenDict(), skip_tables=()) :
+        """
+        copies the pan_dat object into a new, obfuscated, simplified pan_dat object
+
+        :param pan_dat: a ticdat object
+
+        :param table_prepends: a dictionary with mapping each table to the prepend it should apply
+                               when its entries are renamed.  A valid table prepend must be all caps and
+                               not end with I. Should be restricted to entity tables (single field primary
+                               that is not a foreign key child)
+
+        :param skip_tables: a listing of entity tables whose single field primary key shouldn't be renamed
+
+        :return: A named tuple with the following components.
+
+         copy : a deep copy of the pan_dat argument, with the single field primary key values
+                renamed to simple "short capital letters followed by numbers" strings.
+
+         renamings : a dictionary matching the new entries to their original (table, primary key value)
+                     this entry can be used to cross reference any diagnostic information gleaned from the
+                     obfusimplified copy to the original names. For example, "P5 has no production"
+                     can easily be recognized as "Product KX12212 has no production".
+        """
+        from ticdat import TicDatFactory # TicDatFactory.obfusimplify will do the heavy lifting here
+        msg = []
+        verify(self.good_pan_dat_object(pan_dat, msg.append),
+               "pan_dat not a good object for this factory : %s"%"\n".join(msg))
+        tic_dat = self.copy_to_tic_dat(pan_dat)
+        tdf = self.clone(clone_factory=TicDatFactory)
+        rtn = tdf.obfusimplify(tic_dat, table_prepends, skip_tables)
+        dat2 = tdf.copy_to_pandas(rtn.copy, reset_index=True)
+        RtnType = clt.namedtuple("ObfusimplifyResults", "copy renamings")
+        return RtnType(self.PanDat(**{t: getattr(dat2, t) for t in self.all_tables}), rtn.renamings)
+
+
     def copy_to_ampl(self, pan_dat, field_renamings = None, excluded_tables = None):
         """
         copies the pan_dat object into a new pan_dat object populated with amplpy.DataFrame objects
