@@ -8,10 +8,12 @@ import uuid
 import os
 from ticdat.utils import freezable_factory, verify, case_space_to_pretty, pd, TicDatError, FrozenDict, all_fields
 from ticdat.utils import all_underscore_replacements, stringish, dictish, containerish, debug_break, faster_df_apply
+from ticdat.utils import safe_apply
 from itertools import product, chain
 from collections import defaultdict
 import datetime
 import inspect
+from io import StringIO
 try:
     import numpy
 except:
@@ -61,9 +63,9 @@ class JsonPanFactory(freezable_factory(object, "_isFrozen")):
         :return:
         """
         self.pan_dat_factory = pan_dat_factory
-        to_json_args = inspect.getfullargspec(pd.DataFrame.to_json).args
-        assert "orient" in to_json_args
-        self._modern_pandas = "index" in to_json_args
+        to_json_args = inspect.getfullargspec(pd.DataFrame.to_json)
+        assert "orient" in to_json_args.args + to_json_args.kwonlyargs
+        self._modern_pandas = "index" in to_json_args.args + to_json_args.kwonlyargs
         self._isFrozen = True
     def create_pan_dat(self, path_or_buf, fill_missing_fields=False, orient='split', **kwargs):
         """
@@ -93,7 +95,7 @@ class JsonPanFactory(freezable_factory(object, "_isFrozen")):
         To address this, you need to either use set_data_type for your
         PanDatFactory, or specify "dtype" in kwargs. (The former is obviously better).
         """
-        if stringish(path_or_buf) and os.path.exists(path_or_buf):
+        if safe_apply(os.path.exists)(path_or_buf):
             verify(os.path.isfile(path_or_buf), "%s appears to be a directory and not a file." % path_or_buf)
             with open(path_or_buf, "r") as f:
                 loaded_dict = json.load(f)
@@ -117,7 +119,7 @@ class JsonPanFactory(freezable_factory(object, "_isFrozen")):
                 kwargs_ = dict(kwargs)
                 if "dtype" not in kwargs_:
                     kwargs_["dtype"] = self.pan_dat_factory._dtypes_for_pandas_read(t)
-                rtn[t] = pd.read_json(json.dumps(loaded_dict[f]), orient=orient, **kwargs_)
+                rtn[t] = pd.read_json(StringIO(json.dumps(loaded_dict[f])), orient=orient, **kwargs_)
             missing_fields = {(t, f) for t in rtn for f in all_fields(self.pan_dat_factory, t)
                               if f not in rtn[t].columns}
             if fill_missing_fields:
@@ -161,10 +163,10 @@ class JsonPanFactory(freezable_factory(object, "_isFrozen")):
         pan_dat = self.pan_dat_factory._pre_write_adjustment(pan_dat)
         jdict = {}
         def fix_cell(x):
-            if isinstance(x, (pd.Timestamp, numpy.datetime64, datetime.datetime)):
-                return str(x)
             if pd.isnull(x):
                 return None
+            if isinstance(x, (pd.Timestamp, numpy.datetime64, datetime.datetime)):
+                return str(x)
             return x
 
         for t, (pks, dfs) in self.pan_dat_factory.schema().items():
@@ -218,8 +220,9 @@ class JsonPanFactory(freezable_factory(object, "_isFrozen")):
         pan_dat = self.pan_dat_factory._pre_write_adjustment(pan_dat)
 
         if self._modern_pandas:
-            # FYI - pandas Exception: ValueError: 'index=False' is only valid when 'orient' is 'split' or 'table'
-            kwargs["index"] = index if orient in ("split", "table") else True
+            # FYI - pandas Exception:
+            # ValueError: 'index=False' is only valid when 'orient' is 'split', 'table', records or 'values'
+            kwargs["index"] = index if orient in ("split", "table", "records", "values") else True
         case_space_table_names = case_space_table_names and \
                                  len(set(self.pan_dat_factory.all_tables)) == \
                                  len(set(map(case_space_to_pretty, self.pan_dat_factory.all_tables)))
@@ -565,9 +568,9 @@ class XlsPanFactory(freezable_factory(object, "_isFrozen")):
                 if f in self.pan_dat_factory.default_values.get(t, {}):
                     rtn[t][f] = self.pan_dat_factory.default_values[t][f]
                     missing_fields.remove((t, f))
+        xl.close()
         verify(not missing_fields,
                "The following are (table, field) pairs missing from the %s file.\n%s" % (xls_file_path, missing_fields))
-        xl.close()
         rtn = _clean_pandat_creator(self.pan_dat_factory, rtn)
         if self.pan_dat_factory.xlsx_trailing_empty_rows == "prune":
             from ticdat.pandatfactory import remove_trailing_all_nan
@@ -616,10 +619,8 @@ class XlsPanFactory(freezable_factory(object, "_isFrozen")):
         case_space_sheet_names = case_space_sheet_names and \
                                  len(set(self.pan_dat_factory.all_tables)) == \
                                  len(set(map(case_space_to_pretty, self.pan_dat_factory.all_tables)))
-        writer = pd.ExcelWriter(file_path)
-        for t in self.pan_dat_factory.all_tables:
-            getattr(pan_dat, t).to_excel(writer,
-                                         (case_space_to_pretty(t) if case_space_sheet_names else t)[:_longest_sheet],
-                                         index=False)
-        writer.save()
-        writer.close()
+        with pd.ExcelWriter(file_path) as writer:
+            for t in self.pan_dat_factory.all_tables:
+                getattr(pan_dat, t).to_excel \
+                    (writer, sheet_name=(case_space_to_pretty(t) if case_space_sheet_names else t)[:_longest_sheet],
+                     index=False)
