@@ -3,12 +3,14 @@ Read/write ticDat objects from PostGres database. Requires the sqlalchemy module
 """
 
 from collections import defaultdict
+import math
 from ticdat.utils import freezable_factory, TicDatError, verify, stringish, FrozenDict, find_duplicates
 from ticdat.utils import create_duplicate_focused_tdf, dictish, numericish, safe_apply
 try:
     import sqlalchemy as sa
+    saxt = sa.text
 except:
-    sa = None
+    sa = saxt = None
 try:
     import psycopg2
 except:
@@ -38,8 +40,8 @@ def _pg_name(name):
     return "".join(rtn)
 
 def _active_fld_tables(engine, schema, active_fld):
-    return {_[0] for _ in engine.execute("SELECT table_name FROM information_schema.columns " +
-            f"WHERE table_schema = '{schema}' and column_name = '{active_fld}'")}
+    return {_[0] for _ in engine.execute(saxt("SELECT table_name FROM information_schema.columns " +
+            f"WHERE table_schema = '{schema}' and column_name = '{active_fld}'"))}
 
 class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
     def __init__(self, tdf):
@@ -64,7 +66,7 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
     def check_tables_fields(self, engine, schema, error_on_missing_table=False):
         '''
         throws a TicDatError if there there isn't a postgres schema in engine with the proper tables and fields.
-        :param engine: has an .execute method
+        :param engine: has an .execute method. LEGACY NAME. With modern sqlalchemy, pass a connection
         :param schema: string that represents a postgres schema
         :param error_on_missing_table: boolean - should an error be thrown for missing tables? If falsey, then
                print a warning instead.
@@ -72,15 +74,17 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
                  error_on_missing_table is truthy.
         '''
         tdf = self.tdf
-        verify(schema in [row[0] for row in engine.execute("select schema_name from information_schema.schemata")],
+        verify(schema in [row[0] for row in engine.execute(saxt
+                ("select schema_name from information_schema.schemata"))],
                f"Schema {schema} is missing from engine {engine}")
-        pg_tables = [row[0] for row in engine.execute(
-            f"select table_name from information_schema.tables where table_schema ='{schema}'")]
+        pg_tables = [row[0] for row in engine.execute(saxt(
+            f"select table_name from information_schema.tables where table_schema ='{schema}'"))]
         missing_tables = []
         for table in tdf.all_tables:
             if table in pg_tables:
-                pg_fields = [row[0] for row in engine.execute(f"""SELECT column_name FROM information_schema.columns 
-                             WHERE table_schema = '{schema}' AND table_name = '{table}'""")]
+                pg_fields = [row[0] for row in engine.execute(saxt(
+                    f"""SELECT column_name FROM information_schema.columns 
+                             WHERE table_schema = '{schema}' AND table_name = '{table}'"""))]
                 for field in tdf.primary_key_fields.get(table, ()) + \
                              tdf.data_fields.get(table, ()):
                     matches = [f for f in pg_fields if f == _pg_name(field)]
@@ -145,6 +149,8 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
                 return bool(rtn)
             if rtn is None or rtn == "":
                 return "NULL"
+            if numericish(rtn) and abs(rtn) == float("inf") and numericish(self.tdf.infinity_io_flag):
+                return math.copysign(self.tdf.infinity_io_flag, rtn)
             if stringish(rtn) and rtn:
                 return f"'{rtn}'"
             return rtn
@@ -158,6 +164,8 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
             return fld_type.nullable
 
         def default_sql_str(t, f):
+            if forced_field_types.get((t, f)) == "text" and stringish(self.tdf.default_values[t][f]):
+                return f" DEFAULT {db_default(t, f)}"
             fld_type = self.tdf.data_types.get(t, {}).get(f)
             if (fld_type and fld_type.datetime) or get_fld_type(t, f, '') == "bytea":
                 return ""
@@ -182,7 +190,7 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
 
     def write_schema(self, engine, schema, forced_field_types=None, include_ancillary_info=True):
         """
-        :param engine: typically a sqlalchemy database engine with drivertype postgres (really just needs an .execute)
+        :param engine: has an .execute method. LEGACY NAME. With modern sqlalchemy, pass a connection
 
         :param schema: a string naming the postgres schema to populate (will create if needed)
 
@@ -205,9 +213,13 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
         if not include_ancillary_info:
             from ticdat import TicDatFactory
             tdf = TicDatFactory(**{t: [[], pks + dfs] for t, (pks, dfs) in self.tdf.schema().items()})
+            tdf.set_infinity_io_flag(self.tdf.infinity_io_flag)
             for t, dts in self.tdf.data_types.items():
                 for f, dt in dts.items():
                     tdf.set_data_type(t, f, *dt)
+            for t, dfvs in self.tdf.default_values.items():
+                for f, dfv in dfvs.items():
+                    tdf.set_default_value(t, f, dfv)
             forced_field_types_ = {(t, f): "text" for t, (pks, dfs) in self.tdf.schema().items() for f in pks
                        if f not in tdf.data_types.get(t, {})}
             forced_field_types_.update(forced_field_types)
@@ -216,10 +228,11 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
         verify(not getattr(self.tdf, "generic_tables", None),
                "TicDat for postgres does not yet support generic tables")
 
-        if schema not in [row[0] for row in engine.execute("select schema_name from information_schema.schemata")]:
+        if schema not in [rw[0] for rw in engine.execute(saxt("select schema_name from information_schema.schemata"))]:
             engine.execute(sa.schema.CreateSchema(schema))
         for str in self._get_schema_sql(self.tdf.all_tables, schema, forced_field_types):
-            engine.execute(str)
+            engine.execute(saxt(str))
+        engine.commit() if hasattr(engine, "commit") else None
 
     def _handle_prexisting_rows(self, engine, schema, pre_existing_rows):
         verify(isinstance(pre_existing_rows, dict), "pre_existing_rows needs to dict")
@@ -230,10 +243,13 @@ class _PostgresFactory(freezable_factory(object, "_isFrozen"),):
         for t in reversed(self._ordered_tables()):
             if pre_existing_rows[t] == "delete":
                 try:
-                    engine.execute(f"truncate table {schema}.{t}") # postgres truncate will fail on FKs re:less
+                    engine.execute(saxt(f"truncate table {schema}.{t}")) # postgres truncate will fail on FKs re:less
                 except Exception as e:
                     assert "foreign key" in str(e), "truncate should only fail due to foreign key issues"
-                    engine.execute(f"DELETE FROM {schema}.{t}")
+                    engine.rollback() if hasattr(engine, "rollback") else None
+                    engine.execute(saxt(f"DELETE FROM {schema}.{t}"))
+                engine.commit() if hasattr(engine, "commit") else None
+
 
 class PostgresTicFactory(_PostgresFactory):
     """
@@ -284,7 +300,7 @@ class PostgresTicFactory(_PostgresFactory):
         """
         Create a TicDat object from a PostGres connection
 
-        :param engine: A sqlalchemy connection to the PostGres database
+        :param engine: has an .execute method. LEGACY NAME. With modern sqlalchemy, pass a connection
 
         :param schema : The name of the schema to read from
 
@@ -322,8 +338,8 @@ class PostgresTicFactory(_PostgresFactory):
             assert tdf.primary_key_fields.get(table) or tdf.data_fields.get(table), "since no generic tables"
             fields = [_pg_name(f) for f in tdf.primary_key_fields.get(table, ()) +
                       tdf.data_fields.get(table, ())]
-            for row in engine.execute(f"Select {', '.join(fields)} from {schema}.{table}" +
-                                      (f" where {active_fld} is True" if table in active_fld_tables else "")):
+            for row in engine.execute(saxt(f"Select {', '.join(fields)} from {schema}.{table}" +
+                                      (f" where {active_fld} is True" if table in active_fld_tables else ""))):
                 if tdf.primary_key_fields.get(table):
                     pk = [self._read_data_cell(table, f, x) for f,x in
                           zip(tdf.primary_key_fields[table], row[:len(tdf.primary_key_fields[table])])]
@@ -339,7 +355,7 @@ class PostgresTicFactory(_PostgresFactory):
         """
         Find the row counts for duplicated rows.
 
-        :param engine: A sqlalchemy Engine object that can connect to our postgres instance
+        :param engine: has an .execute method. LEGACY NAME. With modern sqlalchemy, pass a connection
 
         :param schema: Name of the schema within the engine's database to use
 
@@ -384,12 +400,12 @@ class PostgresTicFactory(_PostgresFactory):
                 fields = list(map(_pg_name, fields))
                 if t in active_fld_tables:
                     fields.append(active_fld)
-                    datarow = datarow + (True,)
+                    datarow = datarow + (True, )
                 if dump_format == "list":
-                    str = f"INSERT INTO {schema}.{t} ({','.join(fields)}) VALUES ({','.join('%s' for _ in fields)})"
-                    rtn.append((str, datarow))
+                    str = f"INSERT INTO {schema}.{t} ({', '.join(fields)}) VALUES ({', '.join(f':{_}' for _ in fields)})"
+                    rtn.append((str, {f: v for f, v in zip(fields, datarow)}))
                 else:
-                    str = f"INSERT INTO {schema}.{t} ({','.join(fields)}) VALUES %s"
+                    str = f"INSERT INTO {schema}.{t} ({', '.join(fields)}) VALUES %s"
                     rtn[str].append(datarow)
         return tuple(rtn) if dump_format == "list" else dict(rtn)
 
@@ -399,7 +415,7 @@ class PostgresTicFactory(_PostgresFactory):
 
         :param tic_dat: the data object to write
 
-        :param engine: a sqlalchemy database engine with drivertype postgres
+        :param engine:  has an .execute method. LEGACY NAME. With modern sqlalchemy, pass a connection
 
         :param schema: the postgres schema to write to (call self.write_schema explicitly as needed)
 
@@ -416,7 +432,9 @@ class PostgresTicFactory(_PostgresFactory):
         :return:
         """
         verify(sa, "sqalchemy needs to be installed to use this subroutine")
-        verify(engine.name=='postgresql',
+        engine_name = safe_apply(getattr)(engine, "name") or \
+                      safe_apply(getattr)(safe_apply(getattr)(engine, "engine"), "name")
+        verify(engine_name =='postgresql',
                "a sqlalchemy engine with drivername='postgres' is required")
         verify(not dsn or psycopg2, "need psycopg2 to use the faster dsn write option")
         verify(_pg_name(active_fld) ==  active_fld, "active_fld needs to be compliant with PG naming conventions")
@@ -441,7 +459,8 @@ class PostgresTicFactory(_PostgresFactory):
             if len(all_dat) > 1000:
                 print("***pgtd.py not using most efficient data writing technique**")
             for sql_str, data in all_dat:
-                engine.execute(sql_str, data)
+                engine.execute(saxt(sql_str), data)
+        engine.commit() if hasattr(engine, "commit") else None
 
 
 class PostgresPanFactory(_PostgresFactory):
@@ -474,7 +493,7 @@ class PostgresPanFactory(_PostgresFactory):
         """
         Create a PanDat object from a PostGres connection
 
-        :param engine: A sqlalchemy connection to the PostGres database
+        :param engine: has an .execute method. LEGACY NAME. With modern sqlalchemy, pass a connection
 
         :param schema : The name of the schema to read from
 
@@ -504,13 +523,13 @@ class PostgresPanFactory(_PostgresFactory):
         return self.tdf._general_post_read_adjustment(rtn, push_parameters_to_be_valid=True)
 
     def write_data(self, pan_dat, engine, schema, pre_existing_rows=None, active_fld="",
-                   progress=None):
+                   progress=None, table_specific_context_manager=None):
         '''
         write the PanDat data to a postgres database
 
         :param pan_dat: a PanDat object
 
-        :param engine: A sqlalchemy connection to the PostGres database
+        :param engine: has an .execute method. LEGACY NAME. With modern sqlalchemy, pass a connection
 
         :param schema: The postgres schema to write to (call self.write_schema explicitly as needed)
 
@@ -523,9 +542,31 @@ class PostgresPanFactory(_PostgresFactory):
 
         :param progress: if provided, a ticdat.Progress object that is called every time a table is uploaded
 
+        :param table_specific_context_manager: if provided, a dict mapping table name to a context manager factory.
+                That is to say, table_specific_context_manager.values() should be zero argument calleables that return
+                context manager objects. The DataFrame.to_sql statement writing to the database will happen within this
+                resulting context manager for every entry in this dict.
+                This is an expert level only feature - don't use it without studying this whole subroutine
+
         :return:
         '''
         verify(_pg_name(active_fld) ==  active_fld, "active_fld needs to be compliant with PG naming conventions")
+        table_specific_context_manager = table_specific_context_manager or {}
+        verify(isinstance(table_specific_context_manager, dict), "table_specific_context_manager needs to be dict")
+        verify(set(table_specific_context_manager).issubset(self.tdf.all_tables),
+               "bad table_specific_context_manager keys")
+        verify(all(map(callable, table_specific_context_manager.values())),
+               "values of table_specific_context_manager should be context managers (and thus calleable)")
+        verify(all(all(hasattr(v, a) for a in ["__enter__", "__exit__"]) for v in
+                   table_specific_context_manager.values()),
+               "values of table_specific_context_manager should be context managers")
+        class EmptyContextManager(object):
+            def __enter__(self, *execinfo):
+                pass
+            def __exit__(self, *excinfo):
+                pass
+        table_specific_context_manager = {t: table_specific_context_manager.get(t, EmptyContextManager)
+                                          for t in self.tdf.all_tables}
         active_field_tables = _active_fld_tables(engine, schema, active_fld) if active_fld else set()
         self._check_good_pgtd_compatible_table_field_names()
         msg = []
@@ -541,6 +582,8 @@ class PostgresPanFactory(_PostgresFactory):
             df.rename(columns={f: _pg_name(f) for f in fields}, inplace=True)
             if table in active_field_tables:
                 df[active_fld] = True
-            df.to_sql(name=table, schema=schema, con=engine, if_exists="append", index=False)
+            with table_specific_context_manager[table]():
+                df.to_sql(name=table, schema=schema, con=engine, if_exists="append", index=False)
             if progress and not progress.numerical_progress("Uploading...", 100.*i/len(to_upload)):
                 break
+        engine.commit() if hasattr(engine, "commit") else None
