@@ -92,6 +92,16 @@ class TestIO(unittest.TestCase):
             self.assertTrue("TicDatError" in e.__class__.__name__)
             return str(e)
 
+    def test_quickie(self):
+        tdf = TicDatFactory(this=[["Field One"], []])
+        tdf2 = TicDatFactory(this=[["Field One"], ["Field Two"]])
+        dat = tdf.TicDat(this=[["a"], ["b"]])
+        s = self.firesException(lambda :  tdf2.json.create_tic_dat(tdf.json.write_file(dat, "")))
+        self.assertTrue("this cannot be treated as a ticDat table : Inconsistent data row lengths." in s)
+        pdf2 = tdf2.clone(clone_factory=PanDatFactory)
+        s = self.firesException(lambda : pdf2.json.create_pan_dat(tdf.json.write_file(dat, "")))
+        self.assertTrue("this cannot be treated as a PanDat table : insufficient number of columns." in s)
+
     def testXlsSimple(self):
         if not self.can_run:
             return
@@ -165,6 +175,7 @@ class TestIO(unittest.TestCase):
             self.assertFalse(pdf._same_data(dat, dat_1, epsilon=1e-5))
             protein = dat_1.categories["name"] == "protein"
             self.assertTrue(list(dat_1.categories[protein]["maxNutrition"])[0] == 999999999)
+            dat_1.categories['maxNutrition'] = dat_1.categories['maxNutrition'].astype(float)
             dat_1.categories.loc[protein, "maxNutrition"] = float("inf")
             self.assertTrue(pdf._same_data(dat, dat_1, epsilon=1e-5))
 
@@ -204,7 +215,8 @@ class TestIO(unittest.TestCase):
                            ["json", core_path + "2.json", "write_file"],
                            ["xls", core_path+".xlsx", "write_file"]]:
             getattr(getattr(pdf, attr), func)(dat, path)
-            dat_1 = getattr(pdf, attr).create_pan_dat(path)
+            # FutureWarning: Inferring datetime64[ns] from data containing strings is deprecated
+            dat_1 = getattr(pdf, attr).create_pan_dat(path) # happens here, but only when "2.json" in path
             self.assertFalse(pdf._same_data(dat, dat_1))
             self.assertFalse(pdf.find_data_type_failures(dat_1) or pdf.find_data_row_failures(dat_1))
             dat_1 = pdf.copy_to_tic_dat(dat_1)
@@ -221,10 +233,10 @@ class TestIO(unittest.TestCase):
                             {True, False})
 
     def testDateTimeTwo(self):
-        file = os.path.join(_scratchDir, "datetime_pd.xls")
+        file = os.path.join(_scratchDir, "datetime_pd.xlsx")
         df = utils.pd.DataFrame({"a":list(map(utils.pd.Timestamp,
             ["June 13 1960 4:30PM", "Dec 11 1970 1AM", "Sept 11 2001 9:30AM"]))})
-        df.to_excel(file, "Cool Runnings")
+        df.to_excel(file, sheet_name="Cool Runnings")
         pdf = PanDatFactory(cool_runnings = [["a"],[]])
         pdf.set_data_type("cool_runnings", "a", datetime=True)
         dat = pdf.xls.create_pan_dat(file)
@@ -343,7 +355,7 @@ class TestIO(unittest.TestCase):
         ex = self.firesException(lambda : pdf2.json.create_pan_dat(pdf.json.write_file_pd(panDat, "")))
         self.assertTrue("missing" in ex and "extra" in ex)
         ex = self.firesException(lambda : pdf2.json.create_pan_dat(pdf.json.write_file(panDat, "")))
-        self.assertTrue("missing" in ex and "extra" in ex)
+        self.assertTrue("foods cannot be treated as a PanDat table : insufficient number of columns." == ex)
 
         panDat2 = pdf2.sql.create_pan_dat(sqlFilePath, fill_missing_fields=True)
         self.assertTrue(set(panDat2.foods["extra"]) == {0})
@@ -742,6 +754,21 @@ class TestIO(unittest.TestCase):
             dat_1 = utils._get_dat_object(pdf, "create_pan_dat", path, f_or_d, False)
             self.assertTrue(pdf._same_data(dat, dat_1, nans_are_same_for_data_rows=True))
 
+    def testJsonOrientFalseIndex(self):
+        if not self.can_run:
+            return
+
+        ticDat = TicDatFactory(**spacesSchema()).TicDat(**spacesData())
+        panDat = pan_dat_maker(spacesSchema(), ticDat)
+        pdf = PanDatFactory(**spacesSchema())
+
+        for orient in ("split", "table", "records", "values"):
+            with self.subTest(orient=orient):
+                file_path = os.path.join(_scratchDir, f'orient_{orient}.json')
+                pdf.json.write_file_pd(panDat, file_path, orient=orient)
+                panDatRead = pdf.json.create_pan_dat(file_path, orient=orient)
+                self.assertTrue(pdf._same_data(panDat, panDatRead), orient)
+
     def testLongName(self):
         prepend = "b"*20
         pdf = PanDatFactory(**{prepend*2+t:v for t,v in dietSchema().items()})
