@@ -22,6 +22,59 @@ def _deep_anonymize(x)  :
 class TestUtils(unittest.TestCase):
     canRun = False
 
+    def test_i_236(self):
+        schema = PanDatFactory(any_table=[['PK'], ['Column1', 'Column2']])
+        schema.set_data_type(table='any_table', field='PK', strings_allowed=(), number_allowed=True, must_be_int=True)
+        schema.set_data_type(table='any_table', field='Column1', strings_allowed='*', number_allowed=False)
+        schema.set_data_type(table='any_table', field='Column2', strings_allowed='*', number_allowed=False)
+
+        df1 = utils.pd.DataFrame({'PK': [], 'Column1': [], 'Column2': []})
+        df2 = utils.pd.DataFrame({'PK': [1, 2], 'Column1': ['col1_value1', 'col1_value2'],
+                                  'Column2': ['col2_value1', 'col2_value2']})
+        dat1 = schema.PanDat()
+        dat1.any_table = df1.astype({'Column2': 'string'}, copy=True)
+        dat2 = schema.PanDat()
+        dat2.any_table = df1.astype({'Column2': str}, copy=True)
+        dat3 = schema.PanDat()
+        dat3.any_table = df2.astype({'Column2': 'string'}, copy=True)
+        dat4 = schema.PanDat()
+        dat4.any_table = df2.astype({'Column2': str}, copy=True)
+        for dat in [dat1, dat2, dat3, dat4]:
+            self.assertFalse(schema.find_data_type_failures(dat))
+
+    def test_the_wacky_fk_bug_issue_173(self):
+        pdf = PanDatFactory(parent=[[], ["Field"]], child=[[],["Field"]])
+        pdf.add_foreign_key("child", "parent", ["Field", ] * 2)
+        # if you expose it, the crash will look like TypeError: '<' not supported between instances of 'str' and 'int'
+        crashes = lambda _: utils.safe_apply(pdf.find_foreign_key_failures)(_) is None
+        # getting a crash is a funky business, as you can see. it needs duplications
+        dat_1 = pdf.PanDat(child=[["now"], ["now"]], parent=[[1], [2]])
+        dat_2 = pdf.PanDat(child=[["a"], ["a"]], parent=[[1], [2]])
+        dat_3 = pdf.PanDat(child=[["a"], ["b"]], parent=[[1], [2]])
+        # was  [True, True, False] before I fixed it
+        self.assertTrue(list(map(crashes, [dat_1, dat_2, dat_3])) == [False, False, False])
+        fails = {('child', 'parent', ('Field', 'Field')): [True, True]}
+        for d in [dat_1, dat_2, dat_3]:
+             self.assertTrue(pdf.find_foreign_key_failures(d, verbosity="Low", as_table=False) == fails)
+        # its also not symmetrical
+        dat_1 = pdf.PanDat(parent=[["now"], ["now"]], child=[[1], [2]])
+        dat_2 = pdf.PanDat(parent=[["a"], ["a"]], child=[[1], [2]])
+        dat_3 = pdf.PanDat(parent=[["a"], ["b"]], child=[[1], [2]])
+        self.assertTrue(list(map(crashes, [dat_1, dat_2, dat_3])) == [False, False, False])
+        for d in [dat_1, dat_2, dat_3]:
+             self.assertTrue(pdf.find_foreign_key_failures(d, verbosity="Low", as_table=False) == fails)
+
+        # just to make sure nothing is crashing in TicDat world (this makes sense - this is a pandas issue)
+        tdf = pdf.clone(clone_factory=TicDatFactory)
+        crashes = lambda _: utils.safe_apply(tdf.find_foreign_key_failures)(_) is None
+        dat_1 = tdf.TicDat(child=[["now"], ["now"]], parent=[[1], [2]])
+        dat_2 = tdf.TicDat(child=[["a"], ["a"]], parent=[[1], [2]])
+        dat_3 = tdf.TicDat(child=[["a"], ["b"]], parent=[[1], [2]])
+        self.assertTrue(list(map(crashes, [dat_1, dat_2, dat_3])) == [False, False, False])
+        for d in [dat_1, dat_2, dat_3]:
+            fails = tdf.find_foreign_key_failures(d, verbosity="Low")
+            self.assertTrue(len(fails) == 1 and fails[('child', 'parent', ('Field', 'Field'))][1] == (0, 1))
+
     def test_advanced_kwargs_cloning(self):
         if not self.canRun:
             return
@@ -120,14 +173,17 @@ class TestUtils(unittest.TestCase):
         self.assertTrue(pdf.good_pan_dat_object(dat2))
         delattr(dat2, "nodes")
         msg = []
+        self.assertFalse(pdf.good_pan_dat_object(dat2))
         self.assertFalse(pdf.good_pan_dat_object(dat2, msg.append))
         self.assertTrue(msg[-1] == "nodes not an attribute.")
 
         dat3 = pdf.copy_pan_dat(dat)
         dat3.cost.drop("commodity", axis=1, inplace=True)
+        self.assertFalse(pdf.good_pan_dat_object(dat3))
         self.assertFalse(pdf.good_pan_dat_object(dat3, msg.append))
         self.assertTrue("The following are (table, field) pairs missing from the data" in msg[-1])
 
+
         dat4 = pdf.copy_pan_dat(dat)
         dat4.cost["cost"] += 1
         self.assertFalse(pdf._same_data(dat, dat4))
@@ -136,11 +192,21 @@ class TestUtils(unittest.TestCase):
         dat5 = pdf2.copy_pan_dat(dat)
         self.assertTrue(pdf._same_data(dat, dat5))
         self.assertTrue(pdf2._same_data(dat, dat5))
-        dat.commodities = dat.commodities.append(dat.commodities[dat.commodities["name"] == "Pencils"])
-        dat.arcs = dat.arcs.append(dat.arcs[dat.arcs["destination"] == "Boston"])
+        dat.commodities = utils.pd.concat([dat.commodities, dat.commodities[dat.commodities["name"] == "Pencils"]])
+        dat.arcs = utils.pd.concat([dat.arcs, dat.arcs[dat.arcs["destination"] == "Boston"]])
         self.assertFalse(pdf2._same_data(dat, dat5))
         self.assertFalse(pdf._same_data(dat, dat5))
 
+        dat6 = pdf.copy_pan_dat(dat)
+        delattr(dat6, "nodes")
+        dat6.cost.drop("commodity", axis=1, inplace=True)
+        msg = []
+        self.assertFalse(pdf.good_pan_dat_object(dat6))
+        self.assertFalse(pdf.good_pan_dat_object(dat6, msg.append))
+        self.assertTrue(len(msg) == 2)
+        self.assertTrue(any("The following are (table, field) pairs missing from the data" in _ for _ in msg))
+        self.assertTrue("nodes not an attribute." in msg)
+
     def testDataTypes(self):
         if not self.canRun:
             return
@@ -523,6 +589,36 @@ class TestUtils(unittest.TestCase):
         self.assertFalse(input_schema.find_foreign_key_failures(new_pan_dat))
         self.assertTrue(input_schema._same_data(orig_pan_dat, new_pan_dat))
 
+    def testXToManyThree(self):
+        input_schema = PanDatFactory(complex_parent=[["Name", "Value"], ["Data"]],
+                                     simple_parent=[["Name"], ["Data"]],
+                                     child_table=[["Complex", "Simple"], ["Data"]])
+        input_schema.add_foreign_key("child_table", "simple_parent", ["Simple", "Name"])
+        input_schema.add_foreign_key("child_table", "complex_parent", ["Complex", "Name"])
+        self.assertTrue({fk.cardinality for fk in input_schema.foreign_keys} ==
+                        {'many-to-one', 'many-to-many'})
+
+        dat = input_schema.PanDat(complex_parent=[["A", _, _ * 10] for _ in range(1, 5)] +
+                                  [["B", _, _ * 5] for _ in range(2, 5)],
+                                  simple_parent=[["Fred", 100], ["Larry", 200]],
+                                  child_table=[[c, s, i+1] for i, (c, s) in
+                                               enumerate(itertools.product(["A", "B"], ["Fred", "Larry"]))])
+        self.assertFalse(input_schema.find_foreign_key_failures(dat))
+
+        dat2 = input_schema.PanDat(complex_parent=[["A", _, _ * 10] for _ in range(1, 5)],
+                                   simple_parent=dat.simple_parent, child_table=dat.child_table)
+        fails = input_schema.find_foreign_key_failures(dat2, verbosity="Low")
+        self.assertTrue(set(fails) == {('child_table', 'complex_parent', ('Complex', 'Name'))})
+        df = fails['child_table', 'complex_parent', ('Complex', 'Name')]
+        self.assertTrue(len(df) == 2 and set(df["Complex"]) == {'B'})
+
+        dat3 = input_schema.PanDat(complex_parent=dat.complex_parent, child_table=dat.child_table,
+                                   simple_parent=[["Larry", 200]])
+        fails = input_schema.find_foreign_key_failures(dat3, verbosity="Low")
+        self.assertTrue(set(fails) == {('child_table', 'simple_parent', ('Simple', 'Name'))})
+        df = fails['child_table', 'simple_parent', ('Simple', 'Name')]
+        self.assertTrue(len(df) == 2 and set(df["Simple"]) == {'Fred'})
+
     def _testPdfReproduction(self, pdf):
         def _tdfs_same(pdf, pdf2):
             self.assertTrue(pdf.schema() == pdf2.schema())
@@ -867,6 +963,34 @@ class TestUtils(unittest.TestCase):
                          foods = [[f"food_{_}", 10] for _ in range(1, 0)]) # empty list results
         self.assertTrue(dat._len_dict() == {"nutritionQuantities": 10})
 
+
+    def test_obfusimplify(self):
+        # only need a simple test since the heavy work is being done by TicDatFactory
+        pdf = PanDatFactory(products=[["Name"], []], recipes=[["Name"], []],
+                            bill_of_materials=[["Recipe", "Component", "Assembly"], ["Quantity"]],
+                            byproduct=[['Recipe', 'Assembly', 'Byproduct'], ['Quantity']])
+        pdf.add_foreign_key('bill_of_materials', 'products', ('Component', 'Name'))
+        pdf.add_foreign_key('bill_of_materials', 'products', ('Assembly', 'Name'))
+        pdf.add_foreign_key('bill_of_materials', 'recipes', ('Recipe', 'Name'))
+        pdf.add_foreign_key('byproduct', 'bill_of_materials', (('Recipe', 'Recipe'), ('Assembly', 'Assembly')))
+        pdf.add_foreign_key('byproduct', 'products', ('Byproduct', 'Name'))
+        dat = pdf.PanDat(products=[["A"], ["B"], ["C"]], recipes=[["hey"], ["now"]],
+                         bill_of_materials=[["hey", "A", "B", 1], ["now", "A", "B", 112],
+                                            ["now", "C", "B", 1.12]],
+                         byproduct=[["hey", "B", "C", 0.12], ["now", "B", "A", 7]])
+        self.assertTrue(not pdf.find_foreign_key_failures(dat))
+        copy, renamings = pdf.obfusimplify(dat)
+        self.assertFalse(pdf.find_foreign_key_failures(copy))
+        self.assertTrue(dat._len_dict() == copy._len_dict())
+        self.assertFalse(set(dat.products["Name"]).intersection(copy.products["Name"]) or
+                         set(dat.recipes["Name"]).intersection(copy.recipes["Name"]))
+        tic_dat = pdf.copy_to_tic_dat(dat)
+        copy_dat = pdf.copy_to_tic_dat(copy)
+        for (r, a, by), row in copy_dat.byproduct.items():
+            self.assertTrue(tic_dat.byproduct[renamings[r][1], renamings[a][1], renamings[by][1]]["Quantity"] ==
+                            row["Quantity"] > 0)
+
+
 # Run the tests.
 if __name__ == "__main__":
     if not DataFrame :
