@@ -243,8 +243,8 @@ class TestXls(unittest.TestCase):
                     for row in data:
                         for f, c in zip(fields, row):
                             d[f].append(c)
-                    utils.pd.DataFrame(d).to_excel(writer, t, index=False)
-                writer.save()
+                    utils.pd.DataFrame(d).to_excel(writer, sheet_name=t, index=False)
+                writer.close()
 
 
 
@@ -589,17 +589,14 @@ class TestXls(unittest.TestCase):
                             for _ in v.values()))
 
     def testDateTimeTwo(self): # this is good test for datetime stuff
-        file = os.path.join(_scratchDir, "datetime_pd.xls")
+        file = os.path.join(_scratchDir, "datetime_pd.xlsx")
         df = utils.pd.DataFrame({"a":list(map(utils.pd.Timestamp,
             ["June 13 1960 4:30PM", "Dec 11 1970 1AM", "Sept 11 2001 9:30AM"]))})
         tdf = TicDatFactory(cool_runnings = [["a"],[]])
         tdf.set_data_type("cool_runnings", "a", datetime=True)
-        df.to_excel(file, "Cool Runnings")
+        df.to_excel(file, sheet_name="Cool Runnings")
         dat = tdf.xls.create_tic_dat(file)
         self.assertTrue(set(dat.cool_runnings) == set(df["a"]))
-        file = file + "x"
-        df.to_excel(file, "Cool Runnings")
-        dat = tdf.xls.create_tic_dat(file)
         for x, y in zip(sorted(dat.cool_runnings), sorted(set(df["a"]))):
             delta = x-y
             self.assertTrue(abs(delta.total_seconds()) < 1e-4)
