@@ -18,9 +18,9 @@ except:
     dateutil = None
 import json
 try:
-    import roundoffconnect
+    import dspotconnect
 except:
-    roundoffconnect = None
+    dspotconnect = None
 try:
     import pandas as pd
     from pandas import DataFrame
@@ -33,18 +33,18 @@ try:
 except:
     drm = None
 import inspect
+import warnings
 
 def faster_df_apply(df, func, trip_wire_check=None):
     """
-    pandas.DataFrame.apply is rarely used because it is slow. It is slow because it creates a Series for each row
-    of the DataFrame, and passes this Series to the function. faster_df_apply creates a dict for each row of the
-    DataFrame instead, and as a result is **much** faster.
+    In earlier versions of, DataFrame.apply(func=func, axis=1) was very slow. faster_df_apply was faster.
+    See https://bit.ly/3xnLFld for details.
 
-    See https://bit.ly/3xnLFld.
+    Currently, (pandas 2.2.2 or newer) it appears that DataFrame.apply(func=func, axis=1) will be faster than
+    faster_df_apply. We're investigating this matter to be sure. Assuming these results hold up, a future release
+    of ticdat will simply be a redirect.
 
-    It's certainly possible newer versions of pandas will implement a more performant DataFrame.apply. The broader
-    point is, row-wise apply should not be discarded wholesale for performance reasons, as DataFrame.itertuples
-    is reasonably fast
+    See https://github.com/ticdat/ticdat/issues/228 for continuing updates.
 
     :param df: a DataFrame
 
@@ -98,8 +98,8 @@ def clone_add_a_column(tdf_pdf, table, field, field_type, field_position="append
     verify(table in tdf_pdf.all_tables, "Unrecognized table name %s" % table)
     verify(isinstance(field, str), "field needs to be a string")
     verify(field_type in ['primary key', 'data'], "field_type needs to be 'primary key' or 'data'")
-    current_fields = getattr(tdf_pdf, {"primary key": "primary_key_fields", "data": "data_fields"}[field_type])
-    verify(field not in current_fields, f"{field} already present in {current_fields}")
+    current_fields = getattr(tdf_pdf, {"primary key": "primary_key_fields", "data": "data_fields"}[field_type])[table]
+    verify(field not in current_fields, f"{field} already present in {table}")
     if field_position == "append":
         field_position = len(current_fields)
     verify(0 <= field_position <= len(current_fields),
@@ -161,6 +161,8 @@ def clone_rename_a_column(tdf_pdf, table, field, new_field):
             return tuple(do_renaming(_) for _ in mp)if containerish(mp) else mp
         if needs_renaming(fk.mapping):
             rtn.add_foreign_key(fk.native_table, fk.foreign_table, do_renaming(fk.mapping))
+    if (table, field) in tdf_pdf.tooltips:
+        rtn.set_tooltip(table, new_field, tdf_pdf.tooltips[table, field])
     return rtn
 
 def clone_a_anchillary_info_schema(schema, table_restrictions, fields_to_remove=None):
@@ -221,8 +223,8 @@ def clone_a_anchillary_info_schema(schema, table_restrictions, fields_to_remove=
         elif k == "parameters":
             rtn[k] = v if k in table_restrictions else {}
         else:
-            assert k in {"infinity_io_flag", "xlsx_trailing_empty_rows", "duplicates_ticdat_init"}, \
-                f"{k} is unexpected part of schema"
+            assert k in {"infinity_io_flag", "xlsx_trailing_empty_rows", "duplicates_ticdat_init",
+                         "automunge_multitype_fields"}, f"{k} is unexpected part of schema"
             rtn[k] = v
     return rtn
 
@@ -231,21 +233,30 @@ def dateutil_adjuster(x):
         return x
     # note that pd.Timestamp tends to create NaT from Falsey, this is ok so long as you check for null using pd.isnull
     # also, pd.Timestampp can do weird things making Timestamps from numbers, so not enabling that.
+    def _repr_safe(x): # see issues 201 and 208
+        if x is not None and safe_apply(repr)(x) is not None:
+            return x
     def _try_to_timestamp(y):
         if pd and not numericish(y):
-            rtn = safe_apply(pd.Timestamp)(y)
+            with warnings.catch_warnings(): # suppressing the warning because if/when such strings
+                warnings.filterwarnings(    # fail going forward, they will just be handled by
+                    "ignore",               # dateutil.parser
+                    category=FutureWarning,
+                    message="Parsing 'PST' as tzlocal.*"
+                )
+                rtn = safe_apply(pd.Timestamp)(y)
             if not pd.isnull(rtn):
                 return rtn
         if dateutil:
             return safe_apply(dateutil.parser.parse)(y)
-    rtn = _try_to_timestamp(x)
+    rtn = _repr_safe(_try_to_timestamp(x))
     if rtn is not None:
         return rtn
     if not numericish(x):
-        return _try_to_timestamp(str(x))
+        return _repr_safe(_try_to_timestamp(str(x)))
 
 def acceptable_default(v) :
-    return numericish(v) or stringish(v) or (v is None)
+    return numericish(v) or stringish(v) or v in [True, False] or v is None
 
 def all_fields(tpdf, tbl):
     assert tbl in tpdf.all_tables
@@ -261,7 +272,10 @@ class TypeDictionary(namedtuple("TypeDictionary",
         if self.datetime:
             # data is not None and dateutil_adjuster(data) is None will always be invalid, re:less of self.nullable
             # self.nullable only refers to whether the true None can be passed.
-            return isinstance(data, datetime_.datetime) or dateutil_adjuster(data) is not None
+            if isinstance(data, datetime_.datetime):
+                return True
+            munged = dateutil_adjuster(data)
+            return munged is not None and safe_apply(repr)(munged) is not None # see issue 201 for details
         if numericish(data):
             if not self.number_allowed:
                 return False
@@ -428,7 +442,8 @@ def standard_main(input_schema, solution_schema, solve, case_space_table_names=F
 
     Implements a command line signature of
 
-    "python engine_file.py --input <input_file_or_dir> --output <output_file_or_dir> --roundoff <roundoff config file>"
+    "python engine_file.py --input <input_file_or_dir> --output <output_file_or_dir> --foresta <Foresta config file>
+                           --errors <error_file_or_dir>"
 
     For the input/output command line arguments.
 
@@ -450,7 +465,7 @@ def standard_main(input_schema, solution_schema, solve, case_space_table_names=F
 
     Defaults are input.xlsx, output.xlsx
 
-    The roundoff config file is optional. See ticdat wiki for a description of the roundoff config file.
+    The Foresta config file is optional. See ticdat wiki for a description of the Foresta config file.
     """
     verify(all(isinstance(_, ticdat.TicDatFactory) for _ in (input_schema, solution_schema)) or
            all(isinstance(_, ticdat.PanDatFactory) for _ in (input_schema, solution_schema)),
@@ -465,14 +480,14 @@ def standard_main(input_schema, solution_schema, solve, case_space_table_names=F
     file_name = sys.argv[0]
     def usage():
         print (f"python {file_name} --help --input <input file or dir> --output <output file or dir> " +
-               "--roundoff <roundoff config file>")
+               "--foresta <Foresta config file> --errors <errors file or dir>")
     try:
-        opts, args = getopt.getopt(sys.argv[1:], "hi:o:r:e:", ["help", "input=", "output=", "roundoff=", "errors="])
+        opts, args = getopt.getopt(sys.argv[1:], "hi:o:f:e:", ["help", "input=", "output=", "foresta=", "errors="])
     except getopt.GetoptError as err:
         print (str(err))  # will print something like "option -a not recognized"
         usage()
         sys.exit(2)
-    input_file, output_file, roundoff_file, error_file = "input.xlsx", "output.xlsx", None, None
+    input_file, output_file, foresta_file, error_file = "input.xlsx", "output.xlsx", None, None
     for o, a in opts:
         if o in ("-h", "--help"):
             usage()
@@ -481,60 +496,60 @@ def standard_main(input_schema, solution_schema, solve, case_space_table_names=F
             input_file = a
         elif o in ("-o", "--output"):
             output_file = a
-        elif o in ("-r", "--roundoff"):
-            roundoff_file = a
+        elif o in ("-f", "--foresta"):
+            foresta_file = a
         elif o in ("-e", "--errors"):
             error_file = a
         else:
             verify(False, "unhandled option")
 
-    if roundoff_file and error_file:
-        print("The -r and -e command line arguments are incompatible. Use one, or the other, or neither, but not both.")
+    if foresta_file and error_file:
+        print("The -f and -e command line arguments are incompatible. Use one, or the other, or neither, but not both.")
 
     recognized_extensions = (".json", ".xls", ".xlsx", ".db")
     if create_routine == "create_tic_dat":
         recognized_extensions += (".sql", ".mdb", ".accdb")
 
-    roundoff_based_solve = None
-    if roundoff_file:
-        verify(roundoffconnect, "roundoffconnect not installed")
-        if not os.path.isfile(roundoff_file):
-            print("%s is not a valid roundoff file"%roundoff_file)
+    foresta_based_solve = None
+    if foresta_file:
+        verify(dspotconnect, "dspotconnect not installed")
+        if not os.path.isfile(foresta_file):
+            print("%s is not a valid Foresta file"%foresta_file)
             return
-        with open(roundoff_file, "r") as f:
-            roundoff_dict = json.load(f)
-        verify(isinstance(roundoff_dict, dict) and set(roundoff_dict).issuperset(["app_id", "server", "token"]),
-               f"{roundoff_file} failed to resolve to a proper json dict")
+        with open(foresta_file, "r") as f:
+            foresta_dict = json.load(f)
+        verify(isinstance(foresta_dict, dict) and set(foresta_dict).issuperset(["app_id", "server", "token"]),
+               f"{foresta_file} failed to resolve to a proper json dict")
         valid_modes = ["upload and solve", "upload only", "download from scenario"]
-        roundoff_mode = roundoff_dict.get("mode", valid_modes[0])
-        verify(roundoff_mode in valid_modes, f"mode entry from {roundoff_file} needs to be one of {valid_modes}")
-        con = roundoffconnect.AppConnect(*(roundoff_dict[_] for _ in ["server", "token", "app_id"]))
-        print(f"Connection established with Roundoff app {con.app_name} on server {roundoff_dict['server']}")
+        foresta_mode = foresta_dict.get("mode", valid_modes[0])
+        verify(foresta_mode in valid_modes, f"mode entry from {foresta_file} needs to be one of {valid_modes}")
+        con = dspotconnect.AppConnect(*(foresta_dict[_] for _ in ["server", "token", "app_id"]))
+        print(f"Connection established with Foresta app {con.app_name} on server {foresta_dict['server']}")
         EngineProxy = namedtuple("EngineProxy", ["input_schema", "solution_schema", "solve"])
         dummy_engine = EngineProxy(input_schema, solution_schema, solve)
-        engine_on_roundoff = roundoffconnect.TicDatConnector(con, dummy_engine)
-        def roundoff_upload(dat):
-            upload_kwargs = {"scenario_name": roundoff_dict["scenario"]} if "scenario" in roundoff_dict else {}
-            new_scenario_id = engine_on_roundoff.upload_input_dat(dat, **upload_kwargs)
-            print(f"Loaded data into Roundoff scenario {con.current_scenarios()[new_scenario_id]}")
+        engine_on_foresta = dspotconnect.TicDatConnector(con, dummy_engine)
+        def foresta_upload(dat):
+            upload_kwargs = {"scenario_name": foresta_dict["scenario"]} if "scenario" in foresta_dict else {}
+            new_scenario_id = engine_on_foresta.upload_input_dat(dat, **upload_kwargs)
+            print(f"Loaded data into Foresta scenario {con.current_scenarios()[new_scenario_id]}")
             return new_scenario_id
-        if roundoff_mode == "upload and solve":
-            def roundoff_based_solve(dat):
-                new_scenario_id = roundoff_upload(dat)
+        if foresta_mode == "upload and solve":
+            def foresta_based_solve(dat):
+                new_scenario_id = foresta_upload(dat)
                 con.launch_solve(new_scenario_id)
                 while con.is_solving_underway(new_scenario_id):
                     time.sleep(1)
-                    print("Solving on Roundoff...")
-                print(f"Downloading solution from Roundoff scenario {con.current_scenarios()[new_scenario_id]}")
-                return engine_on_roundoff.download_solution(new_scenario_id)
-        elif roundoff_mode == "upload only":
-            roundoff_based_solve = lambda dat: roundoff_upload(dat) and None
+                    print("Solving on Foresta...")
+                print(f"Downloading solution from Foresta scenario {con.current_scenarios()[new_scenario_id]}")
+                return engine_on_foresta.download_solution(new_scenario_id)
+        elif foresta_mode == "upload only":
+            foresta_based_solve = lambda dat: foresta_upload(dat) and None
             output_file = None
         else:
-            def roundoff_based_solve():
-                verify("scenario" in roundoff_dict, "'download from scenario' mode requires a 'scenario' entry")
-                print(f"Downloading solution from Roundoff scenario {roundoff_dict['scenario']}")
-                return engine_on_roundoff.download_solution(roundoff_dict["scenario"])
+            def foresta_based_solve():
+                verify("scenario" in foresta_dict, "'download from scenario' mode requires a 'scenario' entry")
+                print(f"Downloading solution from Foresta scenario {foresta_dict['scenario']}")
+                return engine_on_foresta.download_solution(foresta_dict["scenario"])
             input_file = None
 
     file_or_dir = lambda f: "file" if any(f.endswith(_) for _ in recognized_extensions) else "directory"
@@ -572,13 +587,18 @@ def standard_main(input_schema, solution_schema, solve, case_space_table_names=F
             return
 
 
-    if roundoff_based_solve is None:
-        sln = solve(dat)
+    if foresta_based_solve is None:
+        result = solve(dat)
+        if containerish(result) and len(result) == 2:
+            msg, sln = result
+            print("**\n" + msg + "\n**\n")
+        else:
+            sln = result
     elif output_file:
-        sln = roundoff_based_solve(*([dat] if input_file else []))
+        sln = foresta_based_solve(*([dat] if input_file else []))
     else:
-        roundoff_based_solve(dat)
-        print(f"Returning without solving, as requested by the 'upload only' mode from {roundoff_file}")
+        foresta_based_solve(dat)
+        print(f"Returning without solving, as requested by the 'upload only' mode from {foresta_file}")
         return
 
     verify(not (sln is not None and safe_apply(bool)(sln) is None),
@@ -810,8 +830,6 @@ def change_fields_with_reserved_keywords(tdf, reserved_keywords, undo=False):
         if original_field in tdf.data_types.get(table, ()):
             rtn.set_data_type(table, new_field,
                               *(tdf.data_types[table][original_field]))
-    if hasattr(tdf,'opl_prepend'):
-        rtn.opl_prepend = tdf.opl_prepend
     if hasattr(tdf,'ampl_prepend'):
         rtn.ampl_prepend = tdf.ampl_prepend
     return rtn
@@ -921,7 +939,11 @@ def safe_apply(f) :
             return None
     return _rtn
 
-def dictish(x): return all(hasattr(x, _) for _ in
+def dictish(x):
+    '''
+    !WATCH OUT! a pandas.DataFrame qualifies as dictish. Probably a dumb subroutine.
+    '''
+    return all(hasattr(x, _) for _ in
                            ("__getitem__", "keys", "values", "items", "__contains__", "__len__"))
 def stringish(x): return all(hasattr(x, _) for _ in ("lower", "upper", "strip"))
 def containerish(x): return all(hasattr(x, _) for _ in ("__iter__", "__len__", "__contains__")) \
@@ -1005,8 +1027,10 @@ def deep_copy(x):
     :param x: the object to deep copy. should be nested dicts, tuples, lists or TicDatFactory/PanDatFactory
     :return: a deep (and unfrozen, other than tuples) copy of x
     '''
-    if isinstance(x, (tuple, str, bool)) or numericish(x) or x is None:
-        return x # tuples are frozen anyway
+    if isinstance(x, (str, bool)) or ticdat.utils.numericish(x) or x is None:
+        return x
+    if isinstance(x, tuple):
+        return tuple(deep_copy(y) for y in x)
     if isinstance(x, frozenset):
         return frozenset({deep_copy(y) for y in x})
     if isinstance(x, set):
@@ -1218,37 +1242,62 @@ class Progress(object):
              print("%s:%s:%s"%(theme.ljust(30), "{:.5f}".format(lower_bound).ljust(20),
                                "{:.5f}".format(upper_bound)))
         return True
-    def gurobi_call_back_factory(self, theme, model) :
+    def gurobi_call_back_factory(self, theme, model, sense="minimize") :
         """
-        Allow a Gurobi  model to call mip_progress. **Only for minimize**
+        Allow a Gurobi  model to call mip_progress.
         :param theme: string describing the type of MIP solve underway
         :param model: a Gurobi model (or ticdat.Model.core_model)
+        :param sense: Either 'minimize' or 'maximize'
         :return: a call_back function that can be passed to Model.optimize
         """
         verify(gu, "gurobipy is not installed and properly licensed")
+        assert sense in ['minimize', 'maximize']
         def rtn(gu_model, where) :
             assert gu_model is model
             if where == gu.GRB.callback.MIP:
                 ub = model.cbGet(gu.GRB.callback.MIP_OBJBST)
                 lb = model.cbGet(gu.GRB.callback.MIP_OBJBND)
+                lb, ub = (lb, ub) if sense == "minimize" else (ub, lb)
                 keep_going = self.mip_progress(theme, lb, ub)
                 if not keep_going :
                     model.terminate()
         return rtn
-    def add_cplex_listener(self, theme, model):
+    def add_xpress_callback(self, theme, problem, sense="minimize"):
+        """
+        Add a callback function to the xpress problem that is connected to this progress object
+        :param theme: short descriptive string
+        :param problem: xpress..problem() object (or ticdat.Model.core_model)
+        :param sense: Either 'minimize' or 'maximize'
+        :return:
+        """
+        assert sense in ['minimize', 'maximize']
+        def callback(prob, object):
+            assert object is None
+            # assert prob is problem # this fails for some reason!
+            ub = prob.attributes.mipobjval
+            lb = prob.attributes.bestbound
+            lb, ub = (lb, ub) if sense == "minimize" else (ub, lb)
+            keep_going = self.mip_progress(theme, lb, ub)
+            return 0 if keep_going else 1
+        problem.addcbmiplog(callback=callback, data=None, priority=0)
+    def add_cplex_listener(self, theme, model, sense="minimize"):
         '''
-        Allow a CPLEX model to call mip_progress. **Only for minimize**
+        Allow a CPLEX model to call mip_progress.
         :param theme: short descriptive string
         :param model: cplex.Model object (or ticdat.Model.core_model)
+        :param sense: Either 'minimize' or 'maximize'
         :return:
         '''
         verify(cplexprogress, "docplex is not installed")
+        assert sense in ['minimize', 'maximize']
         super_self = self
         class MyListener(cplexprogress.ProgressListener):
             def notify_progress(self, progress_data):
-                # this is assuming a minimization problem.
-                ub = float("inf") if progress_data.current_objective is None else progress_data.current_objective
-                keep_going = super_self.mip_progress(theme, progress_data.best_bound, ub)
+                no_feasible_yet = {"minimize": float("inf"), "maximize": -float("inf")}
+                ub = no_feasible_yet if progress_data.current_objective is None else progress_data.current_objective
+                lb = progress_data.best_bound
+                lb, ub = (lb, ub) if sense == "minimize" else (ub, lb)
+                keep_going = super_self.mip_progress(theme, lb, ub)
                 if not keep_going:
                     self.abort()
         model.add_progress_listener(MyListener())
